/**
 * Copyright 2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.nutch.admin.management;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;


public class FileUtil {

  static Logger LOG = Logger.getLogger(FileUtil.class.getName());

  private static class RunningFileFilter implements java.io.FileFilter {
   
    public boolean accept(File file) {
      String name = file.getName().toLowerCase();
      return name.endsWith("running");
    }
    
  }

  private static class DirectoryFileFilter implements java.io.FileFilter {

    private FileSystem fFileSystem;
   
    public DirectoryFileFilter(FileSystem fileSystem) {
      this.fFileSystem = fileSystem;
    }

    public boolean accept(File file) {
      boolean ret = false;
      try {
        ret = this.fFileSystem.isDirectory(file);
      } catch (IOException e) {
        LOG.warning(e.toString());
      }
      return ret;
    }
  }

  
  public static long size(File folder, Configuration configuration)
      throws IOException {
    
    FileSystem fileSystem = FileSystem.get(configuration);
    File[] files = fileSystem.listFiles(folder);
    long size = 0;
    for (int i = 0; files != null && i < files.length; i++) {
      File file = files[i];
      if (fileSystem.isDirectory(file)) {
        size = size + size(file, configuration);
      }
      size = size + fileSystem.getLength(file);
    }
    return size + fileSystem.getLength(folder);
  }

  /**
   * @return true if fetch.done exists
   */
  public static boolean isFetched(File segment, Configuration configuration)
      throws IOException {
    
    return exists(configuration, segment, "fetch.done");
  }

  /**
   * @return true if invert.done exists
   */
  public static boolean isInverted(File segment, Configuration configuration)
      throws IOException {
    
    return exists(configuration, segment, "invert.done");
  }

  /**
   * @return true if parse.done exists
   */
  public static boolean isParsed(File segment, Configuration configuration)
      throws IOException {
    
    return exists(configuration, segment, "parse.done");
  }

  /**
   * @return true if parse.done exists
   */
  public static boolean isIndexed(File segment, Configuration configuration)
      throws IOException {
    
    FileSystem system = FileSystem.get(configuration);
    File[] files = system.listFiles(new File(segment, "index"));
    boolean ret = false;
    for (int i = 0; i < files.length; i++) {
      //e.g. file = part-00000
      File file = files[i];
      if(file.isDirectory() && file.getName().startsWith("part-")) {
        ret = exists(configuration, file, "index.done");
        if(!ret) {
          break;
        }
      }
    }
    return ret;
  }

  /**
   * @return true if parse.done exists
   */
  public static boolean isInjected(File instanceFolder,
      Configuration configuration) throws IOException {
    
    return exists(configuration, instanceFolder, "crawldb");
  }

  /**
   * @return true if search.done exists
   */
  public static boolean isReadyToSearch(File segment,
      Configuration configuration) throws IOException {
    
    return exists(configuration, segment, "search.done");
  }

  /**
   * @return true if fileName in folder exists
   */
  private static boolean exists(Configuration configuration, File folder,
      String fileName) throws IOException {
    FileSystem fileSystem = FileSystem.get(configuration);
    return fileSystem.exists(new File(folder, fileName));
  }

  /**
   * @return true if parse.done exists
   */
  public static List getRunningFiles(File folder, Configuration configuration)
      throws IOException {
    FileSystem fileSystem = FileSystem.get(configuration);
    File[] files = fileSystem.listFiles(folder, new RunningFileFilter());
    List list = new LinkedList();
    for (int i = 0; i < files.length; i++) {
      File file = files[i];
      list.add(file.getName());
    }
    return list;
  }
  
  /**
   * @return  folders in this folder
   */
  public static File[] listFolders(File folder, 
      Configuration configuration) 
      throws IOException {
    
    FileSystem system = FileSystem.get(configuration);
    return system.listFiles(folder, new DirectoryFileFilter(system));
  }

}
