/**
 * Copyright 2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.nutch.admin.scheduling;

import java.io.File;
import java.io.IOException;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.nutch.admin.GuiConfigUtil;
import org.apache.nutch.crawl.CrawlDb;
import org.apache.nutch.crawl.Generator;
import org.apache.nutch.crawl.LinkDb;
import org.apache.nutch.fetcher.Fetcher;
import org.apache.nutch.indexer.Indexer;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.StatefulJob;

public class AdminCrawl implements StatefulJob {

  private static class RunningFileFilter implements java.io.FileFilter {
    public boolean accept(File file) {
      String name = file.getName().toLowerCase();
      return name.endsWith("running");
    }
  }

  public void execute(JobExecutionContext arg0) throws JobExecutionException {
    JobDataMap jobDataMap = arg0.getJobDetail().getJobDataMap();
    File crawldbFile = (File) jobDataMap.get("crawldb");
    File linkdbFile = (File) jobDataMap.get("linkdb");
    File segments = (File) jobDataMap.get("segments");
    File segment = null;

    File configurationFile = (File) jobDataMap.get("configuration");
    Configuration configuration = 
      GuiConfigUtil.loadNewConfiguration(configurationFile);
    try {

      segment = generateSegment(crawldbFile, segments, configuration);
      fetchSegment(segment, configuration);
      updateCrawldb(crawldbFile, segment, configuration);
      updateLinkDb(linkdbFile, new File[] { segment }, configuration);
      index(
          new File(segment, "index"), 
          crawldbFile, 
          linkdbFile,
          new File[] { segment }, 
          configuration
          );

    } catch (Exception e) {
      throw new JobExecutionException(e.getMessage());
    } finally {
      try {
        FileSystem fileSystem = FileSystem.get(configuration);
        RunningFileFilter filter = new RunningFileFilter();
        deleteFiles(fileSystem.listFiles(crawldbFile, filter), fileSystem);
        deleteFiles(fileSystem.listFiles(linkdbFile, filter), fileSystem);
        deleteFiles(fileSystem.listFiles(segments, filter), fileSystem);
        deleteFiles(fileSystem.listFiles(segment, filter), fileSystem);
        deleteFiles(fileSystem.listFiles(new File(segment, "index"), filter), fileSystem);
      } catch (IOException e) {
        throw new JobExecutionException(e.getMessage());
      }
    }

  }


  private void deleteFiles(File[] files, FileSystem fileSystem)
      throws IOException {
    for (int i = 0; i < files.length; i++) {
      File file = files[i];
      fileSystem.delete(file);
    }

  }

  
  private void index(File index, File crawldbFile, File linkdbFile,
      File[] segments, Configuration configuration) throws IOException {
    FileSystem fileSystem = FileSystem.get(configuration);
    // create running files in segments
    for (int i = 0; i < segments.length; i++) {
      File file = segments[i];
      fileSystem.createNewFile(new File(file, "index.running"));
    }
    // create running files in linkdb
    fileSystem.createNewFile(new File(linkdbFile, "index.running"));
    // create running files in crawldb
    fileSystem.createNewFile(new File(crawldbFile, "index.running"));
    Indexer indexer = new Indexer(configuration);
    indexer.index(index, crawldbFile, linkdbFile, segments);
  }


  private void updateLinkDb(File linkdbFile, File[] segments,
      Configuration configuration) throws IOException {

    FileSystem fileSystem = FileSystem.get(configuration);

    File running = new File(linkdbFile, "linkdb.running");
    // create lock files in segments
    for (int i = 0; i < segments.length; i++) {
      File file = segments[i];
      fileSystem.createNewFile(new File(file, "linkdb.running"));
    }
    // create lockfile in linkdb
    fileSystem.createNewFile(running);

    LinkDb linkDb = new LinkDb(configuration);
    linkDb.invert(linkdbFile, segments);

    for (int i = 0; i < segments.length; i++) {
      File file = segments[i];
      fileSystem.createNewFile(new File(file, "invert.done"));
    }
  }


  private void updateCrawldb(File crawldbFile, File segment,
      Configuration configuration) throws IOException {
    FileSystem fileSystem = FileSystem.get(configuration);
    File runningSegment = new File(segment, "crawldb.running");
    File runningDB = new File(crawldbFile, "crawldb.running");
    fileSystem.createNewFile(runningSegment);
    fileSystem.createNewFile(runningDB);
    CrawlDb crawlDb = new CrawlDb(configuration);
    crawlDb.update(crawldbFile, segment);
  }

 
  private void fetchSegment(File segment, Configuration configuration)
      throws IOException {
    FileSystem fileSystem = FileSystem.get(configuration);
    File running = new File(segment, "fetch.running");
    fileSystem.createNewFile(running);
    Fetcher fetcher = new Fetcher(configuration);
    fetcher.fetch(segment, configuration.getInt("fetcher.threads.fetch", 10),
        configuration.getBoolean("fetcher.parse", true));
    fileSystem.createNewFile(new File(segment, "fetch.done"));
    if (configuration.getBoolean("fetcher.parse", true)) {
      fileSystem.createNewFile(new File(segment, "parse.done"));
    }

  }

  private File generateSegment(File crawldbFile, File segments,
      Configuration configuration) throws IOException {
    FileSystem system = FileSystem.get(configuration);
    File runningGenerateSegment = new File(segments, "generate.running");
    File runningGenerateDB = new File(crawldbFile, "generate.running");
    system.createNewFile(runningGenerateSegment);
    system.createNewFile(runningGenerateDB);
    Generator generator = new Generator(configuration);
    File segment = generator.generate(crawldbFile, segments);
    return segment;
  }

}
